/* globals Plasm */

define(['jquery', 'core/ajax'], ($, Ajax) => {
  const iframeName = 'quilgo-iframe';
  // in seconds
  const defaultExpiresIn = 3_600;
  // for preview mode we only set it 60s
  const previewModeDefaultExpiresIn = 60;
  // in ms
  const defaultPreviewModeShotsInterval = 6_000;
  let submitPreflightAttempt;
  let plasmIframe;
  let plasmIframeLoader;
  let pageWrapper;
  let isPlasmNotAvailable = false;
  let preflightForm;
  let quilgoCheckFinished;
  let cancelPreflightAttempt;
  let PLASM;
  let plasmConfig;
  const collectors = {};
  let cameraAccessTimeout;
  let desyncMS = null;
  let quizData = null;
  let cmid = null;
  let isPlasmPlayed = false;
  let attemptStartTime = null;
  let scheduledStopTimerId = null;
  let isPreviewUser = false;
  let activeTrackWithUnsupportedCollectors = [];
  const behaviors = {
    click: "click",
    dblclick: "dblclick",
    copy: "copy",
    cut: "cut",
    paste: "paste",
    load: "load",
    online: "online",
    offline: "offline",
    blur: "blur",
    focus: "focus",
    beforeunload: "beforeunload",
    contextmenu: "contextmenu",
    change: "change",
    drag: "drag",
    select: "select",
    resize: "resize",
    scroll: "scroll",
    mouseout: "mouseout",
    key: "key",
    unfocus: "unfocus",
  };
  let elementIntervalId;
  let isPreflightFormSubmitted = false;

  const plasmSetConfig = (config) => {
    plasmConfig = config;
  };

  const setQuizData = (quiz) => {
    quizData = quiz;
  };

  const isPlasmCameraEnabled = () => quizData == null ? false : (Number(quizData.plasm_camera) === 1);

  const isPlasmScreenEnabled = () => quizData == null ? false : (Number(quizData.plasm_screen) === 1);

  const isPlasmForceEnabled = () => quizData == null ? false : (Number(quizData.plasm_force) === 1);

  const quizTimeClose = () => {
    if (quizData == null) {
      return null;
    }

    if (quizData.timeclose != null && quizData.timeclose !== 0) {
      return new Date(Number(quizData.timeclose) * 1000).getTime();
    }

    return null;
  }

  const quizTimeLimit = () => {
    if (quizData == null) {
      return null;
    }

    if (quizData.timelimit != null && quizData.timelimit !== 0) {
      return Number(quizData.timelimit);
    }

    return null;
  }

  const syncedDateNow = () => {
    const dms = desyncMS != null ? desyncMS : 0;
    return Date.now() - dms;
  }

  const getReady = async () => {
    if (!PLASM) {
      return;
    }

    await PLASM.getReady().catch(() => {});
  };

  const play = (attemptId) => {
    if (PLASM == null || PLASM.state !== "stopped" || isPlasmPlayed) {
      return;
    }

    const collectibles = Object.values(behaviors);
    const collectQuestionAndOptionEvents = [
      behaviors.copy,
      behaviors.cut,
      behaviors.paste,
      behaviors.click,
      behaviors.dblclick,
      behaviors.change,
      behaviors.select,
    ];

    PLASM.start({
      collectibles: collectibles,
      iframesUsed: true,
      addMetadata: (event) => {
        const meta = { attemptId: attemptId };
        const collectQuestionAndOptionInfo = collectQuestionAndOptionEvents.includes(event.type);
  
        if (collectQuestionAndOptionInfo) {
          const element = event.target;
          const clickedQuestion = element.closest(".que");
          const iframeParentBody = event.srcElement.ownerDocument?.body || null;
  
          if (clickedQuestion) {
            meta.questionId = clickedQuestion.getAttribute("id");
          } else if (iframeParentBody) {
            // Example iframe parent body id q224:1_answer_id
            const iframeParentBodyId = iframeParentBody.getAttribute("data-id");
            if (iframeParentBodyId != null) {
              const questionIdFromIframeParentBody = iframeParentBodyId
                .replace("_answer_id", "")
                .replace(":", "-")
                .replace("q", "");
              
              // add "question-" suffix because question id from ".que" element is "question-224-1"
              meta.questionId = `question-${questionIdFromIframeParentBody}`;
            }
          } 
  
          /**
           * For option id, tt the moment we only collect unique id for input radio type only
           * As per the other options there is no unique id
           * And at the moment in Plasm we no need the option id for patterns 
           */
          const clickedRadio = element.closest('input[type="radio"]');
          if (clickedRadio) {
            meta.optionId = clickedRadio.getAttribute('id');
          }
    
          return meta;
        }
      },
    })
      .then((plasmSessionUuid) => {
        const wsfunction = 'quizaccess_quilgo_store_sessionid';
        const params = {
          'attemptid': attemptId,
          'plasmsessionid': plasmSessionUuid,
          'camera_enabled': Number(quizData.plasm_camera),
          'screen_enabled': Number(quizData.plasm_screen),
          'force_enabled': Number(quizData.plasm_force),
          'cmid': Number(cmid),
        };
        const request = {
          'methodname': wsfunction,
          'args': params
        };

        // eslint-disable-next-line no-unused-expressions
        Ajax.call([request])[0];

        isPlasmPlayed = true;
      })
      .catch(() => {});
  };

  const plasmStop = () => {
    if (!PLASM) {
      return;
    }

    if (scheduledStopTimerId != null) {
      clearTimeout(scheduledStopTimerId);
    }

    PLASM.stop();
  };

  const plasmInit = async () => {
    if (typeof Plasm === "undefined") {
      return false;
    }

    const timeClose = quizTimeClose();
    const timeLimit = quizTimeLimit();
    const isTimeCloseSet = timeClose != null && timeClose > 0;
    const isTimeLimitSet = timeLimit != null && timeLimit > 0;
    let expiresIn = 0;
    
    if (isPreviewUser) {
      expiresIn = previewModeDefaultExpiresIn;
    } else if (isTimeCloseSet || isTimeLimitSet) {
      expiresIn = timeLimit;
      if (isTimeCloseSet) {
        const timeCloseRemainingTime = (timeClose - syncedDateNow()) / 1000;
        expiresIn = expiresIn <= timeCloseRemainingTime ? expiresIn : timeCloseRemainingTime;
      }
    } else {
      expiresIn = defaultExpiresIn;
    }

    try {
      PLASM = new Plasm({
        debug: true,
        clientToken: plasmConfig.clientToken,
        sessionUuid: quizData.plasm_sessionid,
        expiresIn,
        plasmApiUrl: plasmConfig.apiEndpoint,
        useCollectBehavior: true,
      });

      await PLASM.init();

      return true;
    } catch (error) {
      return false;
    }
  };

  const showUnsupportedErrorMessage = () => {
    $(".plasm, .plasm-not-supported").removeClass("quilgo-hidden");
  };

  const getNextSetup = () => {
    let nextSetup;

    // eslint-disable-next-line no-restricted-syntax
    for (const collectorType in collectors) {
      if (collectors[collectorType] === null) {
        nextSetup = collectorType;

        break;
      }
    }

    return nextSetup;
  };

  const adjustCollectorsHints = (useCamera, useScreen) => {
    $(".plasm-hint-camera, .plasm-hint-screen, .plasm-hint-camera-and-screen").addClass("quilgo-hidden");

    if (useCamera) {
      $(".plasm-hint-camera").removeClass("quilgo-hidden");
      $(".consent-title > .plasm-hint-camera").addClass('text-capitalize');
    }

    if (useScreen) {
      $(".plasm-hint-screen").removeClass("quilgo-hidden");
      if (!useCamera) {
        $(".consent-title > .plasm-hint-screen").addClass('text-capitalize');
      }
    }

    if (useCamera && useScreen) {
      $(".plasm-hint-camera-and-screen").removeClass("quilgo-hidden");
    }
  };

  const openConfirmation = () => {
    const cameraStream = collectors.camera && collectors.camera.meta.stream;
    const screenStream = collectors.screen && collectors.screen.meta.stream;

    if (cameraStream) {
      const videoObj = $("#quilgo-video video")[0];

      videoObj.srcObject = cameraStream;
      videoObj.setAttribute("playsinline", true);
      videoObj.play();
    }

    if (screenStream) {
      const videoScreenObj = $("#plasm-screen-video video")[0];

      videoScreenObj.srcObject = screenStream;
      videoScreenObj.setAttribute("playsinline", true);
      videoScreenObj.play();
    }

    adjustCollectorsHints(cameraStream, screenStream);

    if (cameraStream || screenStream) {
      $(".plasm").removeClass("quilgo-hidden");
      $(".plasm-caliber").removeClass("quilgo-hidden");
      quilgoCheckFinished.parent().removeClass('quilgo-hidden');

      quilgoCheckFinished.change(() => {
        if (quilgoCheckFinished.is(':checked')) {
          submitPreflightAttempt.removeClass('quilgo-hidden');
        } else {
          submitPreflightAttempt.addClass('quilgo-hidden');
        }
      });
    }
  };

  const nextSetup = () => {
    const setupNext = getNextSetup();

    if (setupNext === "navigation") {
      // eslint-disable-next-line no-use-before-define
      setup("navigation");
      return;
    }

    if (!setupNext) {
      openConfirmation();

      return;
    }

    $(".plasm-setup").removeClass("quilgo-hidden");
    $(".plasm-setup-collector").addClass("quilgo-hidden");
    $(`.plasm-setup-${setupNext}`).removeClass("quilgo-hidden");
  };

  const setup = (type) => {
    let shotsInterval = defaultPreviewModeShotsInterval;
    if (!isPreviewUser) {
      const { snapshotsMinInterval, screenshotsMinInterval } = PLASM.config;
      const shotsMinInterval = type === "camera" ? snapshotsMinInterval : screenshotsMinInterval;

      const timeLimit = quizTimeLimit();
      const expiresIn = timeLimit != null && timeLimit > 0 ? timeLimit : defaultExpiresIn;
      shotsInterval = Math.max((expiresIn / plasmConfig.maxShotsPerQuiz) * 1000, shotsMinInterval);
    }

    $(".plasm-setup").addClass("quilgo-hidden");
    $(".plasm-consent").remove();
    $(".plasm-error").addClass("quilgo-hidden");

    if (type === "camera") {
      $(".plasm-loading-setup").removeClass("quilgo-hidden");

      cameraAccessTimeout = setTimeout(() => {
        $(".plasm-loading-setup").addClass("quilgo-hidden");
      }, 2000);
    }

    const collectorOptions = {};

    switch (type) {
      case "camera":
      case "screen":
        collectorOptions.interval = shotsInterval;
        break;
      case "navigation":
        collectorOptions.enableUnfocusedShots = true;
        break;
      default:
    }

    PLASM.setupCollector(type, collectorOptions).then((setupResults) => {
      if (type === "camera") {
        clearTimeout(cameraAccessTimeout);
        $(".plasm-loading-setup").addClass("quilgo-hidden");
      }

      switch (setupResults.status) {
        case "permission_denied":
        case "device_not_found":
          $(`.plasm-${type}-access-error`).removeClass("quilgo-hidden");
          break;
        case "not_supported":
          if (isPlasmForceEnabled()) {
            showUnsupportedErrorMessage();
            break;
          }

          nextSetup();
          break;
        case "success":
          if (
            type === "screen" &&
            setupResults.status !== "not_supported" &&
            setupResults.meta.selectedArea !== "monitor"
          ) {
            $(".plasm-screen-area-error").removeClass("quilgo-hidden");
            break;
          }

          collectors[type] = setupResults;
          if (type === "camera") {
            $(".plasm-setup-camera-enabled-hint").removeClass("quilgo-hidden");
          }

          nextSetup();
          break;
        default:
          break;
      }
    });
  };

  const requestConsent = (useCamera, useScreen) => {
    adjustCollectorsHints(useCamera, useScreen);
    $(".plasm").removeClass("quilgo-hidden");
    $(".plasm-consent").removeClass("quilgo-hidden");
  };

  const setupPermissions = async (requestCamera, requestScreen) => {
    const isInitPlasmSuccessfully = await plasmInit();

    if (isInitPlasmSuccessfully === false) {
      return;
    }

    const plasmSupportedCollectors = PLASM.supportedCollectors;
    const unsupportedCollectors = [];
    let isUnsupportedCameraOrScreenCollector = false;

    const isCameraCollectorIncluded = plasmSupportedCollectors.includes("camera");
    const useCamera = requestCamera && isCameraCollectorIncluded;
    if (requestCamera && isCameraCollectorIncluded === false) {
      isUnsupportedCameraOrScreenCollector = true;
    }

    if (useCamera) {
      collectors.camera = null;
    } else {
      unsupportedCollectors.push("camera");
    }

    const isScreenCollectorIncluded = plasmSupportedCollectors.includes("screen");
    const useScreen = requestScreen && isScreenCollectorIncluded;
    if (requestScreen && isScreenCollectorIncluded === false) {
      isUnsupportedCameraOrScreenCollector = true;
    }

    if (useScreen) {
      collectors.screen = null;
    } else {
      unsupportedCollectors.push("screen");
    }

    collectors.navigation = null;

    if (isUnsupportedCameraOrScreenCollector && isPlasmForceEnabled()) {
      showUnsupportedErrorMessage();
      return;
    }

    activeTrackWithUnsupportedCollectors = unsupportedCollectors;
    $(".plasm-setup-camera, .plasm-setup-camera-retry").on("click", (e) => {
      e.preventDefault();
      setup("camera");
    });
    $(".plasm-setup-screen, .plasm-setup-screen-retry").on("click", (e) => {
      e.preventDefault();
      setup("screen");
    });

    requestConsent(useCamera, useScreen);
    nextSetup();
  };

  const onIframeUrlChanged = (iframe, callback) => {
    let lastDispatched = null;

    const dispatchChange = () => {
      const newHref = iframe.contentWindow.location.href;
      plasmIframeLoader.removeClass('quilgo-hidden');

      if (newHref !== lastDispatched) {
        callback(newHref);
        lastDispatched = newHref;
      }
    };

    const unloadHandler = () => {
      setTimeout(dispatchChange, 0);
    };

    const attachUnload = () => {
      iframe.contentWindow.removeEventListener("unload", unloadHandler);
      iframe.contentWindow.addEventListener("unload", unloadHandler);
    }

    iframe.addEventListener("load", () => {
      attachUnload();
      dispatchChange();

      setTimeout(() => {
        plasmIframeLoader.addClass('quilgo-hidden');
      }, 100);
    });

    attachUnload();
  };

  const addIframe = () => {
    const iframe = document.createElement('iframe');
    iframe.setAttribute('name', iframeName);
    iframe.setAttribute('id', iframeName);
    iframe.setAttribute('class', `${iframeName} quilgo-hidden`);
    document.body.append(iframe);
    
    plasmIframe = $(`#${iframeName}`);
  };

  const schedulePlasmStop = () => {
    let timeLimit = previewModeDefaultExpiresIn;
    if (!isPreviewUser) {
      const qTimeLimit = quizTimeLimit();
      timeLimit = qTimeLimit != null && qTimeLimit > 0 ? qTimeLimit : defaultExpiresIn;
    }

    const startTime = attemptStartTime != null
      ? new Date(Number(attemptStartTime) * 1000).getTime()
      : Date.now();

    const timeClose = quizTimeClose();
    const timeLimitEnd = timeClose != null && timeClose > 0 ? timeClose : (startTime + timeLimit * 1000);

    const delay = timeLimitEnd - syncedDateNow();
    if (scheduledStopTimerId !== undefined) {
      clearTimeout(scheduledStopTimerId);
    }

    scheduledStopTimerId = window.setTimeout(() => {
      plasmStop();
    }, delay);
  }

  const getUrlFileLocation = (url) => {
    // Get .php file with location for example host.com/mod/quiz/attempt.php?x=1&y=2 to attempt.php
    const urlObject = new URL(url);

    return urlObject.pathname.split("/").pop();
  }

  const addIframeEvent = () => {
    onIframeUrlChanged(document.getElementById(iframeName), async (newUrl) => {
      const fileLocation = getUrlFileLocation(newUrl);
      const urlLocationIframeShouldRun = ["attempt.php", "summary.php"];
      const iframeShouldRun = urlLocationIframeShouldRun.includes(fileLocation);

      if (iframeShouldRun) {
        const params = (new URL(newUrl)).searchParams;
        const attemptId = params.get('attempt');
        
        const plasmShouldRun = PLASM.state !== "started";
        if (plasmShouldRun) {
          await getReady();
          play(attemptId);
          schedulePlasmStop();
        }
      }

      if (iframeShouldRun === false) {
        plasmStop();
        window.location.href = newUrl;
      }
    });
  };

  const addPageLoader = () => {
    const loaderName = 'quilgo-iframe-loader';
    const loaderContainer = document.createElement('div');
    loaderContainer.setAttribute('id', loaderName);
    loaderContainer.setAttribute('class', `${loaderName} quilgo-hidden`);
    document.body.append(loaderContainer);

    const loaderSpinner = document.createElement('div');
    loaderSpinner.setAttribute('class', 'spinner-border text-primary');
    
    plasmIframeLoader = $(`#${loaderName}`);
    plasmIframeLoader.append(loaderSpinner);
  }

  const initElements = () => {
    quilgoCheckFinished = $('#id_quilgocheckfinished');

    addIframe();
    addPageLoader();
    submitPreflightAttempt = $('#id_submitbutton');
    submitPreflightAttempt.addClass('btn-lg');
    cancelPreflightAttempt = $('#id_cancel');
    cancelPreflightAttempt.addClass('btn-lg');
    pageWrapper = $('#page-wrapper');

    /**
     * Always init check finished as unchecked in case user redirected
     * to the startattempt page due other preflight check failed (e.g wrong password)
     */
    quilgoCheckFinished.attr('checked', false);
  };

  const addSubmitPreflightEvent = () => {
    if (isPlasmNotAvailable) {
      return;
    }

    /**
     * Set start attempt button to the center
     */
    const formActionsContainer = preflightForm.getElementsByClassName('felement');
    if (formActionsContainer.length > 0) {
      const buttonContainer = formActionsContainer[formActionsContainer.length - 1].getElementsByClassName('d-flex');
      if (buttonContainer.length > 0) {
        buttonContainer[0].classList.add('justify-content-center');
      }
    }

    preflightForm.addEventListener("submit", (e) => {
      e.preventDefault();

      if (isPreflightFormSubmitted) {
        return;
      }
      isPreflightFormSubmitted = true;

      $.ajax({
        url: e.target.action,
        method: "POST",
        data: $('#mod_quiz_preflight_form').serialize(),
      })
      .done((res) => {
        const parsedPreflightForm = $($.parseHTML(res)).find('form[id="mod_quiz_preflight_form"]');
        if (parsedPreflightForm.length === 1) {
          preflightForm.submit();

          return;
        }

        plasmIframeLoader.removeClass('quilgo-hidden');
        preflightForm.setAttribute('target', iframeName);
        preflightForm.submit();

        addIframeEvent();

        plasmIframe.removeClass('quilgo-hidden');
        cancelPreflightAttempt.click();
        
        pageWrapper.remove();
      })
    });
  }

  const addDefaultStartFormEvent = () => {
    $('.moodle-dialogue-base').css('display', 'none');
    $('.quizstartbuttondiv [type=submit]').on('click', (e) => {
      e.preventDefault();

      plasmIframeLoader.removeClass('quilgo-hidden');
      pageWrapper.remove();

      if (!isPlasmNotAvailable) {
        preflightForm.setAttribute('target', iframeName);
      }
      preflightForm.submit();

      if (!isPlasmNotAvailable) {
        addIframeEvent();
        plasmIframe.removeClass('quilgo-hidden');
      }
    });
  }

  const waitForElement = (selector, callback) => {
    elementIntervalId = setInterval(() => {
      const targetElement = document.querySelector(selector);
      if (targetElement) {
        clearInterval(elementIntervalId);
        callback(targetElement);
      }
    }, 500);
  };

  return {
    init: async (_plasmConfig, _quiz, _serverTime, _attemptStartTime, _cmid, _isPreviewUser) => {
      const dateNow = Date.now();
      desyncMS = dateNow - new Date(_serverTime).getTime();

      attemptStartTime = _attemptStartTime;
      initElements();

      cmid = _cmid;
      isPreviewUser = _isPreviewUser;
      setQuizData(_quiz);
      plasmSetConfig(_plasmConfig);
      const requestCamera = isPlasmCameraEnabled();
      const requestScreen = isPlasmScreenEnabled();
      isPlasmNotAvailable = typeof Plasm === "undefined";
      await setupPermissions(requestCamera, requestScreen);

      const isShouldNotAskConsent = (requestCamera === false && requestScreen === false) || activeTrackWithUnsupportedCollectors.length === 2;

      if (isPlasmNotAvailable || isShouldNotAskConsent) {
        quilgoCheckFinished.parent().addClass('quilgo-hidden');
        quilgoCheckFinished.attr('checked', true);
      } else {
        cancelPreflightAttempt.addClass('quilgo-hidden');
        submitPreflightAttempt.addClass('quilgo-hidden');
        quilgoCheckFinished.parent().addClass('quilgo-hidden');
      }

      waitForElement("#mod_quiz_preflight_form", (targetElement) => {
        // eslint-disable-next-line prefer-destructuring
        preflightForm = targetElement;
        const formGroupRows = preflightForm.getElementsByClassName('form-group row');

        /**
         * TODO: refactor the logic
         * 
         * NOTE formGroupRows.length = 2 mean only quilgocheckfinished row and button submit row exists in preflight popup elements
         * If only those 2 elements exists, we override the logic to skip the shown of popup to use regular start attempt form submit
         */
        if (formGroupRows.length <= 2 && (isShouldNotAskConsent || isPlasmNotAvailable)) {
          addDefaultStartFormEvent();
        } else {
          addSubmitPreflightEvent();
        }

        /**
         * Override url when user on startattempt.php to avoid user directly to attempt.php
         * without setup tacking. So in this case user will reloaded to view.php
         */
        const isOnStartattempt = getUrlFileLocation(window.location.href) === "startattempt.php";
        if (isOnStartattempt) {
          window.history.replaceState({}, "", `/mod/quiz/view.php?id=${cmid}`);
        }
      });
    },
  }
});
